# local imports:
from ..utils import Constants, Color

# standard libraries:
from tkinter import TclError, EventType, Widget, Event, Misc
from typing import Any, Optional


class HoverMisc(Misc):
    PALETTE_SIZE: int = 12
    PALETTE_UPDATE_MILLISECONDS: int = 6

    @staticmethod
    def is_hover_background(master: Misc, hover_background: str) -> bool:
        try:
            master.winfo_rgb(color=hover_background)
            return True
        except TclError:
            if hover_background in (Constants.AUTO, Constants.DISABLED):
                return True
            return False

    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            hoverbackground: Optional[str] = Constants.AUTO,
            **options: Any) -> None:
        assert isinstance(self, Widget), "This object must be an instance of Widget class."
        hover_background: Optional[str] = options.pop("hoverbg", hoverbackground)
        super().__init__(master, **options)
        self._hover_background: Optional[str] = Constants.DISABLED
        self._palette_colors: list[str] = []
        self._palette_index: int = 0
        self._palette_index_update: Optional[str] = None
        self.bind(sequence="<<UpdatePalette>>", func=self.update_palette, add=True)
        self.configure(hoverbackground=hover_background)
        self.bind(sequence="<Enter>", func=self.on_mouse_interact, add=True)
        self.bind(sequence="<Leave>", func=self.on_mouse_interact, add=True)

    def on_mouse_interact(self, event: Event) -> None:
        if self._palette_index_update is not None:
            self.after_cancel(id=self._palette_index_update)
            self._palette_index_update = None
        is_mouse_hover: bool = event.type == EventType.Enter
        palette_index: int = self._palette_index + (1 if is_mouse_hover else -1)
        palette_index = min(len(self._palette_colors)-1, max(0, palette_index))
        if palette_index == self._palette_index:
            return None
        self._palette_index = palette_index
        self.after(ms=0, func=self.update_background)
        self._palette_index_update = self.after(self.PALETTE_UPDATE_MILLISECONDS, self.on_mouse_interact, event)

    def update_background(self) -> None:
        if not self._palette_colors or self._hover_background == Constants.DISABLED:
            return None
        background: str = self._palette_colors[self._palette_index]
        Misc.configure(self, background=background)
        self.event_generate(sequence="<<UpdateForeground>>", when="now")

    def update_palette(self, event: Optional[Event] = None) -> None:
        if self._hover_background == Constants.DISABLED:
            return None
        background: str = Misc.cget(self, key="background") \
            if event is not None or not self._palette_colors else \
            self._palette_colors[0]
        background = Color.grant_hex_code(self, hex_code=background)
        hover_background: str = self._hover_background
        if hover_background == Constants.AUTO:
            brightness_factor: float = 1.6 if Color.is_darker(hex_code=background) else .6
            hover_background = Color.adjust_brightness(hex_code=background, factor=brightness_factor)
        hover_background = Color.grant_hex_code(self, hex_code=hover_background)
        self._palette_colors = Color.generate_palette(
            start_hex=background,
            end_hex=hover_background,
            num_colors=self.PALETTE_SIZE)
        self.after(ms=0, func=self.update_background)

    def configure(self, **options: Any) -> Any:
        hover_background: Optional[str] = options.pop("hoverbackground", None)
        hover_background = options.pop("hoverbg", hover_background)
        if hover_background is not None and not self.is_hover_background(self, hover_background):
            error_message: str = "unknown color name \"{}\"".format(hover_background)
            raise TclError(error_message)
        result: Any = super().configure(**options)
        if hover_background is not None:
            self._hover_background = hover_background
        background: Optional[str] = options.get("background", None)
        background = options.get("bg", background)
        if background is not None:
            self.event_generate(sequence="<<UpdatePalette>>", when="now")
        elif hover_background is not None:
            self.after(ms=0, func=self.update_palette)
        return result
    config = configure

    def cget(self, key: str) -> Any:
        return self._hover_background if key in Constants.HOVER_BACKGROUND_KEYS else super().cget(key)
    __getitem__ = cget



# local imports:
from ..models import TransparentMisc
from ..utils import Constants

# standard libraries:
from tkinter import Toplevel, Event, Misc, Wm, Tk
from ctypes import windll, WinDLL, byref
from ctypes.wintypes import POINT, RECT
from typing import Any, Optional
from enum import IntEnum


class WindowIndexes(IntEnum):
    WINDOW_STYLE: int = -16
    EXTENDED_WINDOW_STYLE: int = -20


class WindowStyles(IntEnum):
    MAXIMIZE: int = 0x00010000
    MINIMIZE: int = 0x00020000
    RESIZE: int = 0x00040000


class ExtendedWindowStyles(IntEnum):
    TASKBAR: int = 0x00040000


class MotionUtils:
    def __init__(self, master: "WindowUtils") -> None:
        assert isinstance(master, WindowUtils), "The parent master must be an instance of WindowUtils."
        self.master: WindowUtils = master
        self.start_x: Optional[int] = None
        self.start_y: Optional[int] = None

    def start_move(self, event: Event) -> None:
        current_state: str = self.master.master.wm_state()
        x_screen: int = event.x_root
        y_screen: int = event.y_root
        x_window: int = self.master.master.winfo_rootx()
        y_window: int = self.master.master.winfo_rooty()
        self.master.master.wm_state(newstate="normal")
        self.start_x = (x_screen - x_window) % self.master.master.winfo_width()
        self.start_y = (y_screen - y_window) % self.master.master.winfo_height()
        self.master.master.wm_state(newstate=current_state)

    def on_move(self, event: Event) -> None:
        cursor_position: POINT = POINT()
        cursor_position_handler: int = byref(cursor_position)
        self.master._windows_api.GetCursorPos(cursor_position_handler)
        window_handle: int = self.master.retrieve_window_handle()
        self.master._windows_api.ScreenToClient(window_handle, cursor_position_handler)
        window_rectangle: RECT = RECT()
        window_rectangle_handler: int = byref(window_rectangle)
        self.master._windows_api.GetWindowRect(window_handle, window_rectangle_handler)
        new_y: int = self.master.master.winfo_y()+(cursor_position.y-self.start_y)
        if new_y <= 0:
            if self.master.master.wm_state() == "normal":
                self.master.master.wm_state(newstate="zoomed")
            return None
        elif self.master.master.wm_state() == "zoomed":
            self.master.master.wm_state(newstate="normal")
            self.on_move(event)
            return None
        new_x: int = self.master.master.winfo_x()+(cursor_position.x-self.start_x)
        width: int = window_rectangle.right - window_rectangle.left
        height: int = window_rectangle.bottom - window_rectangle.top
        self.master._windows_api.SetWindowPos(window_handle, None, new_x, new_y, width, height, 0)

    def stop_move(self, event: Event) -> None:
        self.start_x = self.start_y = None

    def toggle_maximize(self, event: Event) -> None:
        new_state: str = "zoomed" if self.master.master.wm_state() == "normal" else "normal"
        self.master.master.wm_state(new_state)


class WindowUtils:
    _graphics_api: WinDLL = windll.gdi32
    _windows_api: WinDLL = windll.user32

    def __init__(self, master: "EnhancedWindow") -> None:
        assert isinstance(master, EnhancedWindow), "This object must be an instance of EnhancedWindow."
        self.master: EnhancedWindow = master
        self._motion_utils: MotionUtils = MotionUtils(self)
        self._updated_size: tuple[int, int] = 0, 0
        self._updated_styles: dict[int, int] = {}

    def _on_update_size(self, event: Event) -> None:
        if self._updated_size == (event.width, event.height):
            return None
        self._updated_size = event.width, event.height
        if event.widget != self.master or not self.master.wm_overrideredirect():
            return None
        window_id: int = self.retrieve_window_handle()
        is_width_resizable, is_height_resizable = self.master.wm_resizable()
        start_x: int = 6 if is_width_resizable else 7
        start_y: int = 6 if is_height_resizable else 7
        end_x: int = event.width + (8 if is_width_resizable else 7)
        if self.master.wm_state() == "zoomed":
            end_x -= 1
        end_y: int = event.height + (8 if is_height_resizable else 7)
        new_region: int = self._graphics_api.CreateRectRgn(start_x, start_y, end_x, end_y)
        self._windows_api.SetWindowRgn(window_id, new_region, True)

    def bind_motion(self, widget: Misc) -> None:
        widget.bind(sequence="<Button-1>", func=self._motion_utils.start_move, add=True)
        widget.bind(sequence="<Button1-Motion>", func=self._motion_utils.on_move, add=True)
        widget.bind(sequence="<ButtonRelease-1>", func=self._motion_utils.stop_move, add=True)
        widget.bind(sequence="<Double-Button-1>", func=self._motion_utils.toggle_maximize, add=True)

    def retrieve_window_handle(self) -> int:
        self.master.update_idletasks()
        window_id: int = self.master.winfo_id()
        window_handle: int = self._windows_api.GetParent(window_id)
        return window_handle

    def center_window(self, width: int, height: int) -> None:
        screen_width: int = self.master.winfo_screenwidth()
        screen_height: int = self.master.winfo_screenheight()
        center_x: int = (screen_width - width)//2
        center_y: int = (screen_height - height)//2
        new_geometry: str = "{}x{}+{}+{}".format(width, height, center_x, center_y)
        self.master.wm_geometry(newGeometry=new_geometry)

    def hide_titlebar(self) -> None:
        if self.master.wm_overrideredirect():
            return None
        self.master.wm_overrideredirect(boolean=True)
        window_handle: int = self.retrieve_window_handle()
        self._updated_styles[WindowIndexes.WINDOW_STYLE] = self._windows_api.GetWindowLongPtrW(
            window_handle,
            WindowIndexes.WINDOW_STYLE)
        self._updated_styles[WindowIndexes.EXTENDED_WINDOW_STYLE] = self._windows_api.GetWindowLongPtrW(
            window_handle,
            WindowIndexes.EXTENDED_WINDOW_STYLE)
        new_styles: dict[int, int] = {
            WindowIndexes.WINDOW_STYLE: WindowStyles.MAXIMIZE | WindowStyles.MINIMIZE | WindowStyles.RESIZE,
            WindowIndexes.EXTENDED_WINDOW_STYLE: ExtendedWindowStyles.TASKBAR}
        for index, style in new_styles.items():
            self._windows_api.SetWindowLongPtrW(window_handle, index, style)
        is_window_shown: bool = self.master.wm_state() != "withdrawn"
        if is_window_shown:
            self.master.wm_deiconify()

    def show_titlebar(self) -> None:
        if not self.master.wm_overrideredirect():
            return None
        window_handle: int = self.retrieve_window_handle()
        for index, style in self._updated_styles.items():
            self._windows_api.SetWindowLongPtrW(window_handle, index, style)
        self._updated_styles.clear()
        self.master.wm_overrideredirect(boolean=False)

    def minimize_window(self) -> None:
        is_titlebar_hidden: bool = self.master.wm_overrideredirect()
        if is_titlebar_hidden:
            window_handle: int = self.retrieve_window_handle()
            self._windows_api.ShowWindow(window_handle, 6)
            return None
        self.master.wm_iconify()
    minimize = minimize_window


class EnhancedWindow(Misc, Wm):
    def __init__(self, **options: Any) -> None:
        super().__init__(**options)
        self.utils: WindowUtils = WindowUtils(self)
        self.bind(sequence="<Configure>", func=self.utils._on_update_size, add=True)

    def configure(self, **options: Any) -> Any:
        result: Any = super().configure(**options)
        if Constants.BACKGROUND_KEYS & options.keys():
            def update_children() -> None: TransparentMisc.update_children(self)
            self.after(ms=0, func=update_children)
        return result
    config = configure


class EnhancedTk(EnhancedWindow, Tk):
    def __init__(self, *, className: str = "TkEnhanced", **options: Any) -> None:
        super().__init__(className=className, **options)


class EnhancedToplevel(EnhancedWindow, Toplevel):
    def __init__(self, master: Optional[Misc] = None, *, modal: bool = False, **options: Any) -> None:
        super().__init__(master=master, **options)
        if modal:
            self.grab_set()
            self.master.wait_window(self)
            self.grab_release()



# local imports:
from ..utils import Constants, Color

# standard libraries:
from tkinter import Widget, Event, Misc
from typing import Any, Optional


class ReadableMisc(Misc):
    @staticmethod
    def get_foreground(master: Misc, **options: Any) -> str:
        background: Optional[str] = options.get("background", None)
        background = options.get("bg", background)
        if background is None:
            background = Misc.cget(master, key="background")
        background = Color.grant_hex_code(master, hex_code=background)
        foreground: str = "#ffffff" if Color.is_darker(background) else "#000000"
        return foreground

    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            foreground: Optional[str] = Constants.READABLE,
            **options: Any) -> None:
        assert isinstance(self, Widget), "This object must be an instance of Widget class."
        foreground: Optional[str] = options.pop("fg", foreground)
        super().__init__(master, **options)
        self._foreground: str = Constants.READABLE
        self.configure(foreground=foreground)
        self.bind(sequence="<<UpdateForeground>>", func=self.update_foreground, add=True)

    def update_foreground(self, event: Optional[Event] = None) -> None:
        if self._foreground == Constants.READABLE:
            foreground: str = self.get_foreground(self)
            Misc.configure(self, foreground=foreground)

    def configure(self, **options: Any) -> Any:
        foreground: Optional[str] = options.pop("foreground", None)
        foreground = options.pop("fg", foreground)
        is_readable: bool = foreground == Constants.READABLE
        if is_readable:
            foreground = self.get_foreground(self, **options)
        options["foreground"] = foreground
        result: Any = super().configure(**options)
        if foreground is not None:
            self._foreground = Constants.READABLE if is_readable else foreground
            self.after(ms=0, func=self.update_foreground)
        return result
    config = configure

    def cget(self, key: str) -> Any:
        return self._foreground if key in Constants.FOREGROUND_KEYS else super().cget(key)
    __getitem__ = cget



# local imports:
from ..models import TransparentMisc, ReadableMisc
from ..utils import FontDescription, Constants

# standard libraries:
from tkinter import Event, Label, Misc
from typing import Any, Optional
from tkinter.font import Font


class EnhancedLabel(TransparentMisc, ReadableMisc, Label):
    MINIMUM_PADDING_X: int = 4

    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = Constants.TRANSPARENT,
            foreground: Optional[str] = Constants.READABLE,
            padx: Optional[int] = MINIMUM_PADDING_X,
            text: Optional[str] = "Enhanced Label",
            truncatetext: bool = True,
            **options: Any) -> None:
        super().__init__(master, background=background, foreground=foreground, **options)
        self._text: str = ""
        self._truncate_text: bool = False
        self.configure(padx=padx, text=text, truncatetext=truncatetext)
        self.bind(sequence="<Configure>", func=self.on_configure, add=True)

    def on_configure(self, event: Optional[Event] = None) -> None:
        if not self._truncate_text:
            Misc.configure(self, text=self._text)
            return None
        font: FontDescription = self.cget(key="font")
        font = Font(font=font)
        label_width: int = self.winfo_width()
        minimum_width: int = font.measure(text=Constants.ELLIPSIS)
        if minimum_width > label_width:
            Misc.configure(self, text=Constants.ELLIPSIS)
            return None
        text: str = self._text
        text_width: int = font.measure(text)
        if text_width > label_width:
            text_width += minimum_width
            while text_width > label_width:
                text = text[:-1]
                text_width = font.measure(text=text + Constants.ELLIPSIS)
            text += Constants.ELLIPSIS
        Misc.configure(self, text=text)

    def configure(self, **options: Any) -> Any:
        padding_x: Optional[int] = options.get("padx", None)
        if isinstance(padding_x, int) and padding_x < self.MINIMUM_PADDING_X:
            error_message: str = "Number of \"padx\" must be at least {}.".format(self.MINIMUM_PADDING_X)
            raise ValueError(error_message)
        truncate_text: Optional[bool] = options.pop("truncatetext", None)
        result: Any = super().configure(**options)
        if truncate_text is not None:
            self._truncate_text = bool(truncate_text)
        if "text" in options:
            self._text = Misc.cget(self, key="text")
            self.after(ms=0, func=self.on_configure)
        return result
    config = configure

    def cget(self, key: str) -> Any:
        options: dict[str, Any] = {"text": self._text, "truncatetext": self._truncate_text}
        return options.get(key) if key in options else super().cget(key)
    __getitem__ = cget



# local imports:
from ..models import TransparentMisc

# standard libraries:
from tkinter import PanedWindow, Checkbutton, Scrollbar, Canvas, Scale, Text


class EnhancedPanedWindow(TransparentMisc, PanedWindow):
    ...


class EnhancedCheckbutton(TransparentMisc, Checkbutton):
    ...


class EnhancedScrollbar(TransparentMisc, Scrollbar):
    ...


class EnhancedCanvas(TransparentMisc, Canvas):
    ...


class EnhancedScale(TransparentMisc, Scale):
    ...


class EnhancedText(TransparentMisc, Text):
    ...



# local imports:
from ..models import TransparentMisc
from ..utils import Constants
from . import EnhancedLabel

# standard libraries:
from tkinter import Toplevel, Event, Misc
from typing import Any, Optional


class EnhancedTooltip(Toplevel):
    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            text: Optional[str] = "Enhanced Tooltip",
            **options: Any) -> None:
        super().__init__(master, **options)
        self.setup_tooltip(text)
        self.master.bind(sequence="<Enter>", func=self.show_tooltip, add=True)
        self.master.bind(sequence="<Motion>", func=self.update_position, add=True)
        self.master.bind(sequence="<Leave>", func=self.hide_tooltip, add=True)

    def setup_tooltip(self, text: Optional[str] = None) -> None:
        self.wm_withdraw()
        self.wm_title(string=text)
        self.wm_overrideredirect(boolean=True)
        self.configure(padx=2, pady=2)
        self._label: EnhancedLabel = EnhancedLabel(self, text=text, truncatetext=True)
        self._label.pack_configure(expand=True, fill="both")

    def show_tooltip(self, event: Event) -> None:
        self.update_position(event)
        self.wm_deiconify()

    def update_position(self, event: Event) -> None:
        x_position: int = event.x_root + 6
        y_position: int = event.y_root + 6
        new_position: str = "+{}+{}".format(x_position, y_position)
        self.wm_geometry(newGeometry=new_position)

    def hide_tooltip(self, event: Event) -> None:
        self.wm_withdraw()

    def configure(self, **options: Any) -> Any:
        text: Optional[str] = options.pop("text", None)
        if text is not None:
            self._label.configure(text=text)
        result: Any = super().configure(**options)
        if Constants.BACKGROUND_KEYS & options.keys():
            def update_children() -> None: TransparentMisc.update_children(self)
            self.after(ms=0, func=update_children)
        return result
    config = configure

    def cget(self, key: str) -> Any:
        return self._label.cget(key="text") if key == "text" else super().cget(key)
    __getitem__ = cget



# local imports:
from .types import FontDescription, PathDescription, SizeDescription, ShowScrollbar, Orientation, ResizeMode, Relief
from .constants import Constants
from .color_utils import Color

__all__: tuple[str, ...] = "FontDescription", "PathDescription", "SizeDescription", "ShowScrollbar", "Orientation", "ResizeMode", "Relief", "Constants", "Color"



# local imports:
from ..utils import PathDescription, SizeDescription, ResizeMode, Constants
from ..models import TransparentMisc

# standard libraries:
from tkinter import TclError, Event, Label, Misc
from typing import Any, Optional
from threading import Thread
from os.path import exists

# third-party libraries:
from PIL.Image import Image, open as open_image
from PIL.ImageTk import PhotoImage


class EnhancedImage(TransparentMisc, Label):
    UPDATE_MILLISECONDS: int = 20

    @staticmethod
    def check_options(**options: Any) -> None:
        unknown_options: tuple[str, ...] = "activeforeground", "bitmap", "disabledforeground", "font", "foreground", "takefocus", "text", "textvariable", "underline", "wraplength"
        for option in unknown_options:
            if option not in options:
                continue
            error_message: str = "unknown option \"{}\"".format(option)
            raise TclError(error_message)

    @staticmethod
    def get_image_size(enhanced_image: "EnhancedImage", image: Image) -> tuple[int, int]:
        container_width: int = enhanced_image.winfo_width()
        container_height: int = enhanced_image.winfo_height()
        image_width, image_height = image.size
        if enhanced_image._resize_mode == Constants.CONTAIN:
            scaling_factor: int = min(container_width / image_width, container_height / image_height)
        elif enhanced_image._resize_mode == Constants.COVER:
            scaling_factor: int = max(container_width / image_width, container_height / image_height)
        new_image_width: int = int(image_width * scaling_factor) if enhanced_image._width == Constants.AUTO else enhanced_image._width
        new_image_height: int = int(image_height * scaling_factor)-4 if enhanced_image._height == Constants.AUTO else enhanced_image._height
        return 1 if new_image_width <= 0 else new_image_width, 1 if new_image_height <= 0 else new_image_height

    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = Constants.TRANSPARENT,
            height: Optional[SizeDescription] = Constants.AUTO,
            image: Optional[PathDescription] = None,
            resizemode: ResizeMode = Constants.CONTAIN,
            width: Optional[SizeDescription] = Constants.AUTO,
            **options: Any) -> None:
        """
        Valid option names:
        activebackground, anchor, background, borderwidth, cursor, height, highlightbackground, highlightcolor,
        highlightthickness, image, justify, padx, pady, relief, resizemode, state, width.
        """
        self.check_options(**options)
        super().__init__(master, background=background, **options)
        self._height: SizeDescription = Constants.AUTO
        self._image: Optional[Image] = None
        self._image_path: Optional[PathDescription] = None
        self._photo_image: Optional[PhotoImage] = None
        self._resize_mode: ResizeMode = Constants.CONTAIN
        self._scheduled_update: Optional[str] = None
        self._width: SizeDescription = Constants.AUTO
        self.configure(height=height, image=image, resizemode=resizemode, width=width)
        self.bind(sequence="<Configure>", func=self.schedule_update, add=True)

    def schedule_update(self, event: Optional[Event] = None, milliseconds: int = UPDATE_MILLISECONDS) -> None:
        if self._scheduled_update is not None:
            self.after_cancel(id=self._scheduled_update)
            self._scheduled_update = None
        resize_thread: Thread = Thread(target=self.resize_image, daemon=True)
        self._scheduled_update = self.after(ms=milliseconds, func=resize_thread.start)

    def resize_image(self) -> None:
        if self._image is not None:
            new_size: tuple[int, int] = self.get_image_size(self, image=self._image)
            resized_image: Image = self._image.resize(size=new_size)
            self.update_image(image=resized_image)
        self._scheduled_update = None

    def update_image(self, image: Image) -> None:
        photo_image: PhotoImage = PhotoImage(image)
        if photo_image == self._photo_image:
            return None
        Misc.configure(self, image=photo_image)
        self._photo_image = photo_image

    def configure(self, **options: Any) -> Any:
        self.check_options(**options)
        resize_mode: Optional[ResizeMode] = options.pop("resizemode", None)
        if resize_mode is not None and resize_mode not in Constants.RESIZE_VALUES:
            error_message: str = "Invalid resize mode: \"{}\"".format(resize_mode)
            raise TclError(error_message)
        image: Optional[Image] = None
        image_path: Optional[PathDescription] = options.pop("image", None)
        if image_path:
            if not exists(path=image_path):
                error_message: str = "Image not found: \"{}\"".format(image_path)
                raise TclError(error_message)
            image = open_image(fp=image_path)
        height: Optional[SizeDescription] = options.pop("height", None)
        if height not in (None, Constants.AUTO) and not isinstance(height, int):
            error_message:str="expected \"auto\" or integer but got \"{}\"".format(height)
            raise TclError(error_message)
        width: Optional[SizeDescription] = options.pop("width", None)
        if width not in (None, Constants.AUTO) and not isinstance(width, int):
            error_message:str="expected \"auto\" or integer but got \"{}\"".format(width)
            raise TclError(error_message)
        result: Any = super().configure(**options)
        if width is not None:
            self._width = width
        if height is not None:
            self._height = height
        if image_path == "":
            self._image = self._image_path = self._photo_image = None
        elif image_path is not None:
            self._image = image
            self._image_path = image_path
            self.schedule_update(milliseconds=0)
        if resize_mode is not None:
            self._resize_mode = resize_mode
        return result
    config = configure

    def cget(self, key: str) -> Any:
        self.check_options(key=None)
        options: dict[str, Any] = {
            "height": self._height,
            "image": self._image_path,
            "resizemode": self._resize_mode,
            "width": self._width}
        return options.get(key) if key in options else super().cget(key)
    __getitem__ = cget



# local imports:
from ..models import TransparentMisc, ReadableMisc, HoverMisc
from ..utils import Constants, Color

# standard libraries:
from tkinter import Button, Event, Misc
from typing import Any, Optional


class EnhancedButton(TransparentMisc, ReadableMisc, HoverMisc, Button):
    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = Constants.TRANSPARENT,
            borderwidth: Optional[int] = 1,
            foreground: Optional[str] = Constants.READABLE,
            highlightthickness: Optional[int] = 5,
            hoverbackground: Optional[str] = Constants.AUTO,
            text: Optional[str] = "Enhanced Button",
            **options: Any) -> None:
        super().__init__(
            master,
            background=background,
            borderwidth=borderwidth,
            foreground=foreground,
            highlightthickness=highlightthickness,
            hoverbackground=hoverbackground,
            text=text,
            **options)

    def update_palette(self, event: Optional[Event] = None) -> None:
        if self._hover_background == Constants.DISABLED:
            return None
        background: str = Misc.cget(self, key="background") \
            if event is not None or not self._palette_colors else \
            self._palette_colors[0]
        background = Color.grant_hex_code(self, hex_code=background)
        hover_background: str = self._hover_background
        if hover_background == Constants.AUTO:
            active_background: str = Misc.cget(self, key="activebackground")
            active_background = Color.grant_hex_code(self, hex_code=active_background)
            brightness_factor: float = 1.2 if Color.is_darker(hex_code=active_background) else .8
            hover_background = Color.adjust_brightness(hex_code=active_background, factor=brightness_factor)
        hover_background = Color.grant_hex_code(self, hex_code=hover_background)
        self._palette_colors = Color.generate_palette(
            start_hex=background,
            end_hex=hover_background,
            num_colors=self.PALETTE_SIZE)
        self.after(ms=0, func=self.update_background)

    def configure(self, **options: Any) -> Any:
        result: Any = super().configure(**options)
        if "activebackground" in options:
            self.after(ms=0, func=self.update_palette)
        return result
    config = configure



# standard libraries:
from typing import TypeAlias, Literal, Union, Any
from tkinter.font import Font
from _tkinter import Tcl_Obj
from pathlib import Path

FontDescription: TypeAlias = Union[
    str
    | Font
    | list[Any]
    | tuple[str]
    | tuple[str, int]
    | tuple[str, int, str]
    | tuple[str, int, list[str] | tuple[str, ...]]
    | Tcl_Obj]
PathDescription: TypeAlias = Union[Path, str]
SizeDescription: TypeAlias = Union[str, int]
Relief: TypeAlias = Literal["raised", "sunken", "flat", "ridge", "solid", "groove"]
ResizeMode: TypeAlias = Literal["cover", "contain"]
ShowScrollbar: TypeAlias = Literal["always", "auto", "never"]
Orientation: TypeAlias = Literal["horizontal", "vertical"]



# local imports:
from ..models import TransparentMisc, HoverMisc
from ..utils import Constants

# standard libraries:
from typing import Any, Optional
from tkinter import Frame, Misc


class EnhancedFrame(TransparentMisc, HoverMisc, Frame):
    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = Constants.TRANSPARENT,
            hoverbackground: Optional[str] = Constants.DISABLED,
            **options: Any) -> None:
        super().__init__(master, background=background, hoverbackground=hoverbackground, **options)



# local imports:
from ..models import TransparentMisc, ReadableMisc
from ..utils import Relief, Constants, Color

# standard libraries:
from tkinter import Entry, Event, Misc
from typing import Any, Optional


class EnhancedEntry(TransparentMisc, ReadableMisc, Entry):
    @staticmethod
    def get_placeholder_color(enhanced_entry: "EnhancedEntry") -> str:
        if enhanced_entry._placeholder_color != Constants.AUTO:
            return enhanced_entry._placeholder_color
        background: str = Misc.cget(enhanced_entry, key="background")
        background = Color.grant_hex_code(enhanced_entry, hex_code=background)
        return "#cccccc" if Color.is_darker(hex_code=background) else "#aaaaaa"

    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = Constants.TRANSPARENT,
            borderwidth: int = 6,
            highlightthickness: Optional[int] = 1,
            placeholdercolor: Optional[str] = Constants.AUTO,
            placeholdertext: str = "",
            relief: Optional[Relief] = Constants.FLAT,
            **options: Any) -> None:
        super().__init__(
            master,
            background=background,
            borderwidth=borderwidth,
            highlightthickness=highlightthickness,
            relief=relief,
            **options)
        self._default_color: Optional[str] = None
        self._default_show: str = ""
        self._placeholder_text: str = ""
        self._placeholder_color: str = Constants.AUTO
        self.configure(placeholdertext=placeholdertext, placeholdercolor=placeholdercolor)
        self.bind(sequence="<FocusIn>", func=self.on_focus, add=True)
        self.bind(sequence="<FocusOut>", func=self.on_blur, add=True)

    def update_foreground(self, event: Optional[Event] = None) -> None:
        placeholder_color: str = self.get_placeholder_color(self)
        if Misc.cget(self, key="foreground") == placeholder_color:
            return None
        super().update_foreground(event)

    def on_focus(self, event: Optional[Event] = None) -> None:
        placeholder_color: str = self.get_placeholder_color(self)
        if Misc.cget(self, key="foreground") != placeholder_color:
            return None
        self.delete(first="0", last="end")
        Misc.configure(self, foreground=self._default_color, show=self._default_show)

    def on_blur(self, event: Optional[Event] = None) -> None:
        if Entry.get(self):
            return None
        placeholder_color: str = self.get_placeholder_color(self)
        Misc.configure(self, foreground=placeholder_color, show="")
        self.insert(index="0", string=self._placeholder_text)

    def configure(self, **options: Any) -> Any:
        if hasattr(self, "_placeholder_color"):
            self.on_focus(event=None)
        placeholder_color: Optional[str] = options.pop("placeholdercolor", None)
        if placeholder_color not in (None, Constants.AUTO):
            self.winfo_rgb(color=placeholder_color)
        placeholder_text: Optional[str] = options.pop("placeholdertext", None)
        result: Any = super().configure(**options)
        if placeholder_text is not None:
            self._placeholder_text = str(placeholder_text)
        if placeholder_color is not None:
            self._placeholder_color = placeholder_color
        self._default_color = Misc.cget(self, key="foreground")
        self._default_show = Misc.cget(self, key="show")
        if not self.focus_get():
            self.after(ms=0, func=self.on_blur)
        return result
    config = configure

    def cget(self, key: str) -> Any:
        options: dict[str, Any] = {
            "placeholdercolor": self._placeholder_color,
            "placeholdertext": self._placeholder_text}
        return options.get(key) if key in options else super().cget(key)
    __getitem__ = cget

    def set(self, value: str) -> None:
        self.on_focus(event=None)
        self.delete(first="0", last="end")
        self.insert(index="0", string=value)
        if not self.focus_get():
            self.on_blur(event=None)

    def get(self) -> str:
        placeholder_color: str = self.get_placeholder_color(self)
        return "" if Misc.cget(self, key="foreground") == placeholder_color else super().get()



# local imports:
from .transparent_misc import TransparentMisc
from .readable_misc import ReadableMisc
from .hover_misc import HoverMisc

__all__: tuple[str, ...] = "TransparentMisc", "ReadableMisc", "HoverMisc"



# local imports:
from . import EnhancedScrollbar, EnhancedCanvas, EnhancedFrame
from ..utils import ShowScrollbar, Orientation, Constants

# standard libraries:
from typing import Any, Callable, Optional
from tkinter import TclError, Event, Misc


class EnhancedScrollableFrame(EnhancedFrame):
    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = Constants.TRANSPARENT,
            showscrollbar: ShowScrollbar = Constants.AUTO,
            orient: Orientation = Constants.VERTICAL,
            **options: Any) -> None:
        super().__init__(master, background=background, **options)
        self._binded_widgets: set[Misc] = set()
        self._show_scrollbar: ShowScrollbar = Constants.AUTO
        self._orientation: Orientation = Constants.VERTICAL
        self.setup_widgets()
        self.configure(showscrollbar=showscrollbar, orient=orient)
        self.bind(sequence="<Configure>", func=self.move_children, add=True)

    def move_children(self, event: Event) -> None:
        for child_widget in self.winfo_children():
            if child_widget in (self._scrollbar, self._canvas):
                continue
            self.bind_mouse_wheel(widget=child_widget)
            configure_command: Optional[Callable] = child_widget.grid_configure if child_widget.grid_info() \
                else child_widget.pack_configure if child_widget.pack_info() \
                else child_widget.place_configure if child_widget.place_info() else None
            configure_command(in_=self._frame)
        self.update_scrollbar()

    def bind_mouse_wheel(self, widget: Misc) -> None:
        if widget not in self._binded_widgets:
            widget.bind(sequence="<MouseWheel>", func=self.on_mouse_wheel, add=True)
            self._binded_widgets.add(widget)
        for child_widget in widget.winfo_children():
            self.bind_mouse_wheel(widget=child_widget)

    def setup_widgets(self) -> None:
        self._scrollbar: EnhancedScrollbar = EnhancedScrollbar(self, borderwidth=0, orient="vertical")
        self._canvas: EnhancedCanvas = EnhancedCanvas(
            master=self,
            highlightthickness=0,
            xscrollcommand=self._scrollbar.set,
            yscrollcommand=self._scrollbar.set)
        self._canvas.bind(sequence="<MouseWheel>", func=self.on_mouse_wheel, add=True)
        self._canvas.bind(sequence="<Configure>", func=self.on_configure, add=True)
        self._canvas.pack_configure(expand=True, fill="both")
        self._frame: EnhancedFrame = EnhancedFrame(self._canvas)
        self._frame.bind(sequence="<MouseWheel>", func=self.on_mouse_wheel, add=True)
        self._frame.bind(sequence="<Configure>", func=self.on_configure, add=True)
        self._frame_id: int = self._canvas.create_window(0, 0, anchor="nw", window=self._frame)

    def on_mouse_wheel(self, event: Event) -> None:
        if self._scrollbar.get() == (.0, 1.):
            return None
        scroll_commands: dict[str, Callable] = {
            "vertical": self._canvas.yview_scroll,
            "horizontal": self._canvas.xview_scroll}
        scroll_command: Optional[Callable] = scroll_commands.get(self._orientation, None)
        if scroll_command is None:
            return None
        scroll_amount: int = event.delta // 120
        scroll_command(number=-scroll_amount, what="units")

    def on_configure(self, event: Event) -> None:
        scroll_coordinates: tuple[int, int, int, int] = self._canvas.bbox("all")
        scroll_coordinates: list[int] = list(scroll_coordinates)
        canvas_end_x, canvas_end_y = scroll_coordinates[2:]
        canvas_width, canvas_height = event.width, event.height
        scroll_coordinates[2] = max(canvas_end_x, canvas_width)
        scroll_coordinates[3] = max(canvas_end_y, canvas_height)
        self._canvas.configure(scrollregion=scroll_coordinates)
        is_vertical: bool = self._orientation == "vertical"
        height: int = 0 if is_vertical else canvas_height
        width: int = canvas_width if is_vertical else 0
        self._canvas.itemconfigure(tagOrId=self._frame_id, height=height, width=width)
        self.update_scrollbar()

    def update_scrollbar(self) -> None:
        self._scrollbar.update_idletasks()
        if self._show_scrollbar == "never" or self._show_scrollbar == "auto" and self._scrollbar.get() == (.0, 1.):
            self._scrollbar.pack_forget()
            return None
        sides: dict[str, str] = {"vertical": "right", "horizontal": "bottom"}
        side: Optional[str] = sides.get(self._orientation, None)
        self._scrollbar.pack_configure(before=self._canvas, fill="both", side=side)

    def configure(self, **options: Any) -> Any:
        show_scrollbar: Optional[ShowScrollbar] = options.pop("showscrollbar", None)
        if show_scrollbar not in (None, *Constants.SHOW_VALUES):
            error_message: str = "bad state \"{}\": must be {}, {} or {}".format(
                orientation, *Constants.SHOW_VALUES)
            raise TclError(error_message)
        orientation: Optional[Orientation] = options.pop("orient", None)
        if orientation not in (None, *Constants.ORIENTATION_VALUES):
            error_message: str = "bad orientation \"{}\": must be {} or {}".format(
                orientation, *Constants.ORIENTATION_VALUES)
            raise TclError(error_message)
        result: Any = super().configure(**options)
        if show_scrollbar is not None:
            self._show_scrollbar = show_scrollbar
        if orientation is not None:
            self._orientation = orientation
            commands: dict[str, Callable] = {"vertical": self._canvas.yview, "horizontal": self._canvas.xview}
            command: Optional[Callable] = commands.get(self._orientation, None)
            self._scrollbar.configure(command=command, orient=self._orientation)
        return result
    config = configure

    def cget(self, key: str) -> Any:
        return self._orientation if key == "orient" else super().cget(key)
    __getitem__ = cget



# local imports:
from .enhanced_widgets import EnhancedPanedWindow, EnhancedCheckbutton, EnhancedScrollbar, EnhancedCanvas, EnhancedScale, EnhancedText
from .enhanced_window import EnhancedToplevel, EnhancedTk
from .enhanced_label import EnhancedLabel
from .enhanced_tooltip import EnhancedTooltip
from .enhanced_button import EnhancedButton
from .enhanced_frame import EnhancedFrame
from .enhanced_scrollable_frame import EnhancedScrollableFrame
from .enhanced_entry import EnhancedEntry
from .enhanced_image import EnhancedImage

__all__: list[str] = [
    name
    for name, _ in globals().items()
    if not name.startswith("_") and name.startswith("Enhanced")]



# local imports:
from ..utils import Constants

# standard libraries:
from tkinter import BaseWidget, Event, Misc
from typing import Any, Optional


class TransparentMisc(Misc):
    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = Constants.TRANSPARENT,
            **options: Any) -> None:
        assert isinstance(self, BaseWidget), "This object must be an instance of BaseWidget class."
        background = options.pop("bg", background)
        super().__init__(master, **options)
        self._background: str = Constants.TRANSPARENT
        self.configure(background=background)
        self.bind(sequence="<<ParentUpdate>>", func=self.on_parent_update, add=True)

    def on_parent_update(self, event: Event) -> None:
        if self._background != Constants.TRANSPARENT:
            return None
        parent_background: str = Misc.cget(self.master, key="background")
        if not parent_background:
            return None
        Misc.configure(self, background=parent_background)
        self.event_generate(sequence="<<UpdateForeground>>", when="tail")
        self.event_generate(sequence="<<UpdatePalette>>", when="now")
        self.update_children()

    def update_children(self) -> None:
        for child_widget in self.winfo_children():
            child_widget.event_generate(sequence="<<ParentUpdate>>", when="now")

    def configure(self, **options: Any) -> Any:
        background: Optional[str] = options.pop("background", None)
        background = options.pop("bg", background)
        is_transparent: bool = background == Constants.TRANSPARENT
        if is_transparent:
            background = Misc.cget(self.master, key="background")
        options["background"] = background
        result: Any = super().configure(**options)
        if background is not None:
            self._background = Constants.TRANSPARENT if is_transparent else background
            self.after(ms=0, func=self.update_children)
        return result
    config = configure

    def cget(self, key: str) -> Any:
        return self._background if key in Constants.BACKGROUND_KEYS else super().cget(key)
    __getitem__ = cget



# standard libraries:
from dataclasses import dataclass


@dataclass(frozen=True)
class Constants:
    # Special values for widget settings.
    ELLIPSIS: str = "..."
    AUTO: str = "auto"
    DISABLED: str = "disabled"
    READABLE: str = "readable"
    TRANSPARENT: str = "transparent"
    COVER: str = "cover"
    CONTAIN: str = "contain"
    RESIZE_VALUES: tuple[str, ...] = COVER, CONTAIN
    RAISED: str = "raised"
    SUNKEN: str = "sunken"
    FLAT: str = "flat"
    RIDGE: str = "ridge"
    SOLID: str = "solid"
    GROOVE: str = "groove"
    RELIEF_VALUES: tuple[str, ...] = RAISED, SUNKEN, FLAT, RIDGE, SOLID, GROOVE
    HORIZONTAL: str = "horizontal"
    VERTICAL: str = "vertical"
    ORIENTATION_VALUES: tuple[str, ...] = VERTICAL, HORIZONTAL
    ALWAYS: str = "always"
    NEVER: str = "never"
    SHOW_VALUES: tuple[str, ...] = ALWAYS, AUTO, NEVER

    # Keys for background and foreground options.
    BACKGROUND_KEYS: tuple[str, ...] = "background", "bg"
    FOREGROUND_KEYS: tuple[str, ...] = "foreground", "fg"
    HOVER_BACKGROUND_KEYS: tuple[str, ...] = "hoverbackground", "hoverbg"



# standard libraries:
from struct import unpack
from tkinter import Misc
from re import fullmatch

# third-party libraries:
from numpy import ndarray, linspace


class Color:
    DEFAULT_LUMINANCE_THRESHOLD: float = 128.
    MINIMUM_PALLETE_SIZE: int = 2

    LUMINANCE_RED_COEFFICIENT: float = .2126
    LUMINANCE_GREEN_COEFFICIENT: float = .7152
    LUMINANCE_BLUE_COEFFICIENT: float = .0722

    def __init__(self):
        raise TypeError("This class is not intended to be instantiated.")

    @staticmethod
    def generate_palette(start_hex: str, end_hex: str, num_colors: int = MINIMUM_PALLETE_SIZE) -> list[str]:
        if num_colors < Color.MINIMUM_PALLETE_SIZE:
            error_message: str = "Number of colors must be at least {}.".format(Color.MINIMUM_PALLETE_SIZE)
            raise ValueError(error_message)
        start_color: tuple[int, int, int] = Color.hex_to_rgb(start_hex)
        end_color: tuple[int, int, int] = Color.hex_to_rgb(end_hex)
        colors: list[str] = []
        for index in range(num_colors):
            rgb_values: ndarray = linspace(start=start_color, stop=end_color, num=num_colors, dtype=int)[index]
            hex_color_code: str = Color.rgb_to_hex(*rgb_values)
            colors.append(hex_color_code)
        return colors

    @staticmethod
    def adjust_brightness(hex_code: str, factor: float = 1.) -> str:
        def clamp(value: int) -> int:
            value = max(10, value)
            scaled_value: int = int(value * factor)
            clamped_value: int = max(0, scaled_value)
            return min(255, clamped_value)
        rgb_values: tuple[int, int, int] = Color.hex_to_rgb(hex_code)
        red_adjusted, green_adjusted, blue_adjusted = map(clamp, rgb_values)
        if factor > 1.0:
            red_adjusted = max(red_adjusted, 1)
            green_adjusted = max(green_adjusted, 1)
            blue_adjusted = max(blue_adjusted, 1)
        return Color.rgb_to_hex(red_adjusted, green_adjusted, blue_adjusted)

    @staticmethod
    def grant_hex_code(master: Misc, hex_code: str) -> str:
        if Color.is_hex_code(hex_code):
            return hex_code
        rgb_values: tuple[int, int, int] = master.winfo_rgb(color=hex_code)
        hex_code = Color.rgb_to_hex(*rgb_values)
        return hex_code

    @staticmethod
    def luminance(red_value: int, green_value: int, blue_value: int) -> float:
        return Color.LUMINANCE_RED_COEFFICIENT * red_value + Color.LUMINANCE_GREEN_COEFFICIENT * green_value + Color.LUMINANCE_BLUE_COEFFICIENT * blue_value

    @staticmethod
    def hex_to_rgb(hex_code: str) -> tuple[int, int, int]:
        if not Color.is_hex_code(hex_code):
            error_message: str = "Invalid hex code format: {}".format(hex_code)
            raise ValueError(error_message)
        normalized_hex_code: str = hex_code.lstrip("#")
        if len(normalized_hex_code) == 3:
            normalized_hex_code = "".join(character * 2 for character in normalized_hex_code)
        hex_color_bytes: bytes = bytes.fromhex(normalized_hex_code)
        return unpack("BBB", hex_color_bytes)

    @staticmethod
    def rgb_to_hex(red_value: int, green_value: int, blue_value: int) -> str:
        def clamp(value: int) -> int:
            clamped_value: int = max(0, value)
            return min(255, clamped_value)
        rgb_values: tuple[int, int, int] = red_value, green_value, blue_value
        red_adjusted, green_adjusted, blue_adjusted = map(clamp, rgb_values)
        if not all(0 <= value <= 255 for value in (red_adjusted, green_adjusted, blue_adjusted)):
            raise ValueError("RGB values must be between 0 and 255.")
        return "#{:02X}{:02X}{:02X}".format(red_adjusted, green_adjusted, blue_adjusted)

    @staticmethod
    def is_hex_code(hex_code: str) -> bool:
        return fullmatch(r"#[0-9A-Fa-f]{6}", hex_code) or fullmatch(r"#[0-9A-Fa-f]{3}", hex_code)

    @staticmethod
    def is_darker(hex_code: str, luminance_threshold: float = DEFAULT_LUMINANCE_THRESHOLD) -> bool:
        red_value, green_value, blue_value = Color.hex_to_rgb(hex_code)
        calculated_luminance: float = Color.luminance(red_value, green_value, blue_value)
        return calculated_luminance < luminance_threshold
